#pragma once

namespace JGL {
    enum class Direction : u8 {
        None = 0,
        Vertical = 1,
        Horizontal = 2,
        Diagonal_NWSE = 3, // North West -> South East.
        Diagonal_SWNE = 4 // South West -> North East.
    };

    inline Direction operator|(Direction a, Direction b) {
        return static_cast<Direction>(static_cast<int>(a) | static_cast<int>(b));
    }

    inline bool operator&(Direction a, Direction b) {
        return (u8)a & (u8)b;
    }

    static std::string to_string(const JGL::Direction& direction) {
        switch (direction) {
            case JGL::Direction::None:
                return "None";
            case JGL::Direction::Vertical:
                return "Vertical";
            case JGL::Direction::Horizontal:
                return "Horizontal";
            case JGL::Direction::Diagonal_NWSE:
                return "Diagonal_NWSE";
            case JGL::Direction::Diagonal_SWNE:
                return "Diagonal_SWNE";
            default:
                return "Unknown";
        }
    }

    enum class MSAA_SAMPLE_RATE : u8 {
        MSAA_NONE = 0,
        MSAA_2X = 1,
        MSAA_4X = 2,
        MSAA_8X = 3
    };

    static std::string to_string(const JGL::MSAA_SAMPLE_RATE& sample_rate) {
        switch (sample_rate) {
            case MSAA_SAMPLE_RATE::MSAA_NONE:
                return "No MSAA";
            case MSAA_SAMPLE_RATE::MSAA_2X:
                return "MSAA 2x";
            case MSAA_SAMPLE_RATE::MSAA_4X:
                return "MSAA 4x";
            case MSAA_SAMPLE_RATE::MSAA_8X:
                return "MSAA 8x";
            default:
                return "Unknown";
        }
    }
    static int to_int(const JGL::MSAA_SAMPLE_RATE& sample_rate) {
        switch (sample_rate) {
            case MSAA_SAMPLE_RATE::MSAA_NONE:
                return 0;
            case MSAA_SAMPLE_RATE::MSAA_2X:
                return 2;
            case MSAA_SAMPLE_RATE::MSAA_4X:
                return 4;
            case MSAA_SAMPLE_RATE::MSAA_8X:
                return 8;
            default:
                return 0;
        }
    }
}
