#pragma once

#include <vector>
#include <filesystem>
#include <J3ML/LinearAlgebra.hpp>
#include <filesystem>
#include <iostream>


/// Defines external C references to FreeType Data Structures
extern "C" typedef struct FT_FaceRec_* FT_Face;
extern "C" typedef struct FT_LibraryRec_* FT_Library;

namespace JGL
{

    //bool Init();
    bool InitTextEngine();

    /// A Font class implementation.
    /// Wraps the font's FreeType asset handle and provides helper functions.
    class Font {
    public:
        /// Default constructor does not initialize any members
        Font() = default;
        Font(const std::filesystem::path& path);
        /// Destructor handles freeing of the underlying asset handle.
        ~Font();
        static Font LoadTTF(const std::filesystem::path& filepath);
        static std::vector<Font> GetLoadedFonts();
        /// Returns the bounding-box the given string would occupy at the given point-size, assuming normal (1) scaling.
        /// @param text The string to measure.
        /// @param ptSize The font size at which to measure.
        /// @return The size-in-pixels that would contain the entire text.
        Vector2 MeasureString(const std::string& text, unsigned int ptSize);
    public:
        int index = 0;
        FT_Face face;
    };
}