#pragma once

#include <glad/glad.h>
#include <vector>
#include <array>
#include <unordered_map>


/// TODO: FontCache mechanism works amazing, but makes no fucking sense
/// Let's document and reorganize it to be a little nicer on the mental :)

namespace JGL {
    class CachedGlyph;
    class CachedFont;
    class FontCache;
}

/// Represents a single font-character "glyph", that has been cached in-memory for fast retrieval.
class JGL::CachedGlyph {
private:
    char character;
    std::array<GLfloat, 12> texcoords;
public:
    int x2offset = 0, y2offset = 0, w = 0, h = 0;
    float advanceX = 0, advanceY = 0;

    //CachedGlyph(GLuint texture_id, char c);
    CachedGlyph(char c, std::array<GLfloat, 12> texcoords, float x2o, float y2o, float w, float h, float advX, float advY);
    char getCharacter() const;
    [[nodiscard]] std::array<GLfloat, 12> getTexCoords() const;
};

/// Represents a Font object as it exists in the font-cache.
class JGL::CachedFont {
private:
    std::unordered_map<char, CachedGlyph*> glyphs;
    GLuint texture = 0;
    GLsizei texture_width = 0, texture_height = 0;
    unsigned int font_size = 0;
    unsigned int font_index = 0;
    void Erase();
public:
    void appendGlyph(CachedGlyph* glyph);
    unsigned int getFontSize() const;
    unsigned int getFontIndex() const;
    CachedGlyph* getGlyph(char c);
    std::unordered_map<char, CachedGlyph*> getGlyphs();
    const GLuint* getTextureHandle();
    [[nodiscard]] GLsizei getTextureWidth() const;
    [[nodiscard]] GLsizei getTextureHeight() const;
public:
    CachedFont(GLuint texture_id, GLsizei texture_width, GLsizei texture_height, unsigned int font_size, unsigned int font_index);
    ~CachedFont();
};

class JGL::FontCache {
private:
    std::vector<CachedFont*> cachedFonts{};
public:
    std::vector<CachedFont*> getFonts();
    CachedFont* getFont(unsigned int font_size, unsigned int font_index);
    void appendFont(CachedFont* font);
    void newFont(GLuint texture_id, GLsizei texture_width, GLsizei texture_height, unsigned int font_size, unsigned int font_index);
    void eraseFont(CachedFont* font);
    void purgeCache();
};

namespace JGL {
    inline FontCache fontCache;
}
