#pragma once
#include <ReImage/Image.h>
#include <J3ML/LinearAlgebra.hpp>
#include <Color3.hpp>
#include <Color4.hpp>
#include <glad/glad.h>

namespace JGL {
    using namespace ReImage;
    enum class TextureFilteringMode : u8 {
        NEAREST = 0, //Fastest for 2D, Sometimes causes graphical issues.
        BILINEAR = 1, //Fast and pretty, The best for 2D.

        MIPMAP_NEAREST = 2, //Nearest with mipmaps. The fastest for 3D, Sometimes causes graphical issues. Uses more vram.
        MIPMAP_BILINEAR = 3, //Bilinear with mipmaps, Fast and pretty. Uses more vram.
        MIPMAP_TRILINEAR = 4 //The prettiest. Still decent speed. Uses more vram.
    };

    enum class TextureWrappingMode : u8 {
        REPEAT = 0,
        MIRRORED_REPEAT = 1,
        CLAMP_TO_EDGE = 2,
        CLAMP_TO_BORDER = 3 //Effectively the same as clamp_to_edge
    };

    /// Represents texture data loaded on the GPU. Contains a handle that can be passed to OpenGL draw calls.
    class Texture {
    private:
        void Erase();
    protected:
        GLuint texture_handle = 0;
        Vector2 texture_size = {0, 0};
        ReImage::TextureFlag texture_flags;
        ReImage::TextureFormat texture_format;
        TextureFilteringMode texture_filtering_mode;
        TextureWrappingMode texture_wrapping_mode;
        void load(Image* software_texture, const Vector2& size, const TextureFormat& format, TextureFilteringMode filtering_mode, TextureWrappingMode wrapping_mode);
    public:
        /// Load a texture from a file,
        explicit Texture(const std::string& file, TextureFilteringMode filtering_mode = TextureFilteringMode::BILINEAR, TextureWrappingMode wrapping_mode = TextureWrappingMode::CLAMP_TO_EDGE, const TextureFlag& flags = TextureFlag::INVERT_Y);
        Texture(Image* software_texture, const Vector2& size, const TextureFormat& format, TextureFilteringMode filtering_mode, TextureWrappingMode wrapping_mode);
        /* Initialize a texture filled with trash data
        this is primarily for the RenderTarget */
        explicit Texture(const Vector2& size);
        Texture(const Texture& rhs);
        ~Texture();
    public:
        [[nodiscard]] GLuint GetGLTextureHandle() const;
        [[nodiscard]] Vector2 GetDimensions() const;
        [[nodiscard]] TextureFilteringMode GetFilteringMode() const;
        [[nodiscard]] TextureWrappingMode GetWrappingMode() const;
        [[nodiscard]] TextureFlag GetFlags() const;
        [[nodiscard]] TextureFormat GetFormat() const;
        [[nodiscard]] std::vector<Color4> GetPixelData() const;
    };

}